/* sane - Scanner Access Now Easy.

   pie-usb.c

   Copyright (C) 2011 Michael Rickmann <mrickma@gwdg.de>
   
   This file is part of the SANE package.
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
   
   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.
   
   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.
   
   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.
   
   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.
   
   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice. 
*/



static PIE_USB_Model crystalscan_7200_model = {
  "PIE/Reflecta",		/* Device vendor string */
  "CrystalScan 7200",		/* Device model name */
  0x30,				/* Model ID */
  /* operation mode values for preview, skip calibration and quality */
  {0x00, 0x08, 0x0a},
  75000,			/* default brightness */
  /* R, G, B gain constant */
  {4.19682524E-04, 3.92060196E-04, 3.89647803E-04},
  /* R, G, B normal mode offset factor */
  {1.05, 1.05, 1.05},
  16,				/* minimal gain */
  42,				/* maximal gain in quality mode */
  0x0f00,			/* normal mode maximal exposure time */
  10,				/* gain calibration part */
  {{0x16e6, 0x0ff2, 0x0ff2},	/* RGB texp and gain, first calibration if quality mode */
   0x16e6, {0x21, 0x21, 0x19}},
  {{0x0be2, 0x0bcf, 0x0b88},	/* RGB texp and gain, first calibration if normal mode */
   0x0be2, {0x35, 0x2c, 0x27}}
};


static PIE_USB_Model proscan_7200_model = {
  "PIE/Reflecta",		/* Device vendor string */
  "ProScan 7200",		/* Device model name */
  0x36,				/* Model ID */
  /* operation mode values for preview, skip calibration and quality */
  {0x00, 0x08, 0x02},
  75000,			/* default brightness */
  /* R, G, B gain constant */
  {4.19682524E-04, 3.92060196E-04, 3.89647803E-04},
  /* R, G, B normal mode offset factor */
  {1.2229896394, 1.0447735936, 0.9805181615},
  16,				/* minimal gain */
  42,				/* maximal gain in quality mode */
  0x0f00,			/* normal mode maximal exposure time */
  10,				/* gain calibration part */
  {{0x2c89, 0x1eb7, 0x17ca},	/* RGB texp and gain, first calibration if quality mode */
   0x2c89, {0x25, 0x25, 0x28}},
  {{0x0e79, 0x0bff, 0x0c6c},	/* RGB texp and gain, first calibration if normal mode */
   0x0e79, {0x3f, 0x3d, 0x39}}
};

/* list of USB descriptors, do not mind different models with the same USB id */

static PIE_USB_Device_Entry pie_usb_device_list[] = {
  {0x05e3, 0x0145, &crystalscan_7200_model},	/* Reflecta CrystalScan 7200, id 0x30 */
  {0x05e3, 0x0145, &proscan_7200_model},	/* Reflecta ProScan 7200, id 0x36 */
  {0, 0, 0}
};

/* values for some stereotype USB control write sequences */

static PIE_USB_Value_Data PIE_USB_Init_Sequence_1[] = {
  {VALUE_INIT_1, 0x04},
  {VALUE_INIT_2, 0xff}, {VALUE_INIT_2, 0xaa}, {VALUE_INIT_2, 0x55},
  {VALUE_INIT_2, 0x00},
  {VALUE_INIT_2, 0xff}, {VALUE_INIT_2, 0x87}, {VALUE_INIT_2, 0x78},
  {VALUE_INIT_2, 0x30},
  {VALUE_INIT_1, 0x05}, {VALUE_INIT_1, 0x04},
  {VALUE_INIT_2, 0xff},
  {0x0, 0x0}
};

static PIE_USB_Value_Data PIE_USB_Init_Sequence_2[] = {
  {VALUE_INIT_2, 0xff}, {VALUE_INIT_2, 0xaa}, {VALUE_INIT_2, 0x55},
  {VALUE_INIT_2, 0x00},
  {VALUE_INIT_2, 0xff}, {VALUE_INIT_2, 0x87}, {VALUE_INIT_2, 0x78},
  {VALUE_INIT_2, 0x00},
  {VALUE_INIT_1, 0x05}, {VALUE_INIT_1, 0x04},
  {VALUE_INIT_2, 0xff},
  {0x0, 0x0}
};

static PIE_USB_Value_Data PIE_USB_Setup_SCSI_Sequence[] = {
  {VALUE_INIT_2, 0xff}, {VALUE_INIT_2, 0xaa}, {VALUE_INIT_2, 0x55},
  {VALUE_INIT_2, 0x00},
  {VALUE_INIT_2, 0xff}, {VALUE_INIT_2, 0x87}, {VALUE_INIT_2, 0x78},
  {VALUE_INIT_2, 0xe0},
  {VALUE_INIT_1, 0x05}, {VALUE_INIT_1, 0x04},
  {VALUE_INIT_2, 0xff},
  {0x0, 0x0}
};


/* forward / prototype declarations */

static SANE_Status attach_one (const char *name);
SANE_Status
pie_usb_scsi_wrapper (int fd, const void *src, size_t src_size,
		      void *dst, size_t * dst_size);
static void pie_dump_buffer (int level, unsigned char *buf, int n);

static SANE_Status
sense_handler (int scsi_fd, unsigned char *result, void *arg);


/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ helpers ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static SANE_Status
pie_usb_poke_decints (char *filename, int *nums, int *len)
{
  FILE *afile;
  char buffer[1024];
  char *fres, *nptr;
  char *eptr = NULL;
  long int lnum;
  int i;

  afile = fopen (filename, "r");
  if (afile == NULL)
    {
      DBG (DBG_error, "pie_usb_poke_decints: file %s does not exist\n",
	   filename);
      return SANE_STATUS_IO_ERROR;
    }

  i = 0;
  do
    {
      fres = fgets (buffer, sizeof (buffer), afile);
      if (fres != NULL)
	{
	  nptr = buffer;
	  errno = 0;
	  lnum = strtol (nptr, &eptr, 10);
	  while ((errno == 0) && (nptr != eptr) && (i < *len))
	    {
	      nums[i] = lnum;
	      i++;
	      nptr = eptr;

	      errno = 0;
	      lnum = strtol (nptr, &eptr, 10);
	    }
	}
    }
  while (fres != NULL);
  *len = i;

  fclose (afile);

  return SANE_STATUS_GOOD;
}


static SANE_Status
pie_usb_poke_bytes (char *filename, unsigned char *nums, int *len)
{
  FILE *afile;
  char buffer[1024];
  char *fres, *nptr;
  char *eptr = NULL;
  long int lnum;
  int i;

  afile = fopen (filename, "r");
  if (afile == NULL)
    {
      DBG (DBG_error, "pie_usb_poke_bytes: file %s does not exist\n",
	   filename);
      return SANE_STATUS_IO_ERROR;
    }

  i = 0;
  do
    {
      fres = fgets (buffer, sizeof (buffer), afile);
      if (fres != NULL)
	{
	  nptr = buffer;
	  errno = 0;
	  lnum = strtol (nptr, &eptr, 16);
	  while ((errno == 0) && (nptr != eptr) && (i < *len))
	    {
	      nums[i] = lnum & 0xff;
	      i++;
	      nptr = eptr;

	      errno = 0;
	      lnum = strtol (nptr, &eptr, 16);
	    }
	}
    }
  while (fres != NULL);
  *len = i;

  fclose (afile);

  return SANE_STATUS_GOOD;
}


/* Write data to a pnm file (e.g. calibration). For debugging only */
/* data is RGB or grey, with big endian byte order */
static SANE_Status
pie_usb_write_pnm_file (char *filename, uint8_t * data, int depth,
			int channels, int pixels_per_line, int lines)
{
  FILE *out;
  int count;

  DBG (DBG_info,
       "pie_usb_write_pnm_file: depth=%d, channels=%d, ppl=%d, lines=%d\n",
       depth, channels, pixels_per_line, lines);

  out = fopen (filename, "w");
  if (!out)
    {
      DBG (DBG_error,
	   "pie_usb_write_pnm_file: could nor open %s for writing: %s\n",
	   filename, strerror (errno));
      return SANE_STATUS_INVAL;
    }
  if (depth == 1)
    {
      fprintf (out, "P4\n%d\n%d\n", pixels_per_line, lines);
    }
  else
    {
      fprintf (out, "P%c\n%d\n%d\n%d\n", channels == 1 ? '5' : '6',
	       pixels_per_line, lines, (int) pow (2, depth) - 1);
    }
  if (channels == 3)
    {
      for (count = 0; count < (pixels_per_line * lines * 3); count++)
	{
	  if (depth == 16)
	    fputc (*(data + 1), out);
	  fputc (*(data++), out);
	  if (depth == 16)
	    data++;
	}
    }
  else
    {
      if (depth == 1)
	{
	  pixels_per_line /= 8;
	}
      for (count = 0; count < (pixels_per_line * lines); count++)
	{
	  switch (depth)
	    {
	    case 8:
	      fputc (*(data + count), out);
	      break;
	    case 16:
	      fputc (*(data + 1), out);
	      fputc (*(data), out);
	      data += 2;
	      break;
	    default:
	      fputc (data[count], out);
	      break;
	    }
	}
    }
  fclose (out);

  DBG (DBG_proc, "pie_usb_write_pnm_file: finished\n");
  return SANE_STATUS_GOOD;
}

/* write a set of gray scale pnm files from shading data */
static SANE_Status
pie_usb_shades_to_pnm (Pie_Scanner * scanner, char *name, int lines)
{
  char plane[4][12] = { "-red.pnm", "-green.pnm", "-blue.pnm", "-ired.pnm" };
  char filename[256];
  unsigned char *buffer, *dest;
  int *shade[4];
  int pixels_per_line;
  int snd_length, bits;
  int i, j, k, val;
  SANE_Int status;

  pixels_per_line = scanner->device->cal_info[0].pixels_per_line;
  snd_length = pixels_per_line;
  bits = scanner->device->cal_info[0].receive_bits;
  if (bits > 8)
    snd_length *= 2;

  buffer = malloc (snd_length * lines);
  if (!buffer)
    return SANE_STATUS_NO_MEM;
  for (k = 0; k < 4; k++)
    {
      shade[k] = scanner->cal_data->shades + k * pixels_per_line;
    }

  for (j = 0; j < 4; j++)
    {
      dest = buffer;
      if (bits > 8)
	for (i = 0; i < pixels_per_line; i++)
	  {
	    val = shade[j][i];
	    *dest++ = val & 0xff;
	    *dest++ = (val >> 8) & 0xff;
	  }
      else
	for (i = 0; i < pixels_per_line; i++)
	  *dest++ = shade[j][i] & 0xff;
      for (i = 1; i < lines; i++)
	{
	  memcpy (dest, buffer, snd_length);
	  dest += snd_length;
	}

      strncpy (filename, name, 240);
      strncat (filename, plane[j], 255);
      status = pie_usb_write_pnm_file (filename, buffer, bits,
				       1, pixels_per_line, lines);
      if (status != SANE_STATUS_GOOD)
	return status;
    }

  free (buffer);
  return SANE_STATUS_GOOD;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ low level USB ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ----------------------- PIE_USB_WRITE_CONTROL_SEQUENCE ------------------------ */

static SANE_Status
pie_usb_write_control_sequence (SANE_Int dn, PIE_USB_Value_Data sequ[])
{
  SANE_Int status = SANE_STATUS_GOOD;
  int i;

  DBG (DBG_info, "pie_usb_write_control_sequence writing\n");

  for (i = 0; (status == SANE_STATUS_GOOD) && (sequ[i].bValue != 0); i++)
    {
      status =
	sanei_usb_control_msg (dn, REQUEST_TYPE_OUT, REQUEST_REGISTER,
			       sequ[i].bValue, INDEX, 1, &(sequ[i].bData));
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (DBG_error, "pie_usb_write_control_sequence failed\n");
	  return status;
	}
    }
  return SANE_STATUS_GOOD;
}

/* ---------------------------- PIE_USB_READY_STATE ------------------------------ */

/* Actually, there are two answers, a simple one where 1 byte with 0 or 1 is ok,   */
/* and a more elaborate one in two bytes starting with 3. The second byte signals  */
/* some condition, 0 is ready.                                                     */

static SANE_Status
pie_usb_ready_state (SANE_Int dn)
{
  SANE_Status status;
  SANE_Byte val;

  status =
    sanei_usb_control_msg (dn, REQUEST_TYPE_IN, REQUEST_REGISTER,
			   VALUE_READ_REGISTER, INDEX, 1, &val);
  if (status != SANE_STATUS_GOOD)
    {
      DBG (DBG_error, "pie_usb_ready_state failed at 1st read\n");
      return status;
    }
  DBG (DBG_info, "pie_usb_ready_state got 0x%02x at 1st read\n", val);

  if (val <= 1)
    return SANE_STATUS_GOOD;
  if (val != 3)
    {
      DBG (DBG_error, "pie_usb_ready_state failed\n");
      return SANE_STATUS_INVAL;
    }

  status =
    sanei_usb_control_msg (dn, REQUEST_TYPE_IN, REQUEST_REGISTER,
			   VALUE_READ_REGISTER, INDEX, 1, &val);
  if (status != SANE_STATUS_GOOD)
    {
      DBG (DBG_error, "pie_usb_ready_state failed at 2nd read\n");
      return status;
    }
  DBG (DBG_info, "pie_usb_ready_state got 0x%02x at 2nd read\n", val);

  if (val == 0)
    return SANE_STATUS_GOOD;
  else if (val == 8)
    return SANE_STATUS_DEVICE_BUSY;
  else if (val == 2)
    return SANE_STATUS_IO_ERROR;
  else
    return SANE_STATUS_INVAL;
}


/* --------------------------- PIE_USB_WRITE_SCSI_CMD ---------------------------- */

static SANE_Status
pie_usb_write_scsi_cmd (SANE_Int dn, const SANE_Byte cmnd[6])
{
  SANE_Status status;
  SANE_Byte mnd;
  int i;

  DBG (DBG_proc, "pie_usb_write_scsi_cmd writing 6 bytes\n");

  for (i = 0; i < 6; i++)
    {
      mnd = cmnd[i];
      status =
	sanei_usb_control_msg (dn, REQUEST_TYPE_OUT, REQUEST_REGISTER,
			       VALUE_WRITE_REGISTER, INDEX, 1, &mnd);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (DBG_error, "pie_usb_write_scsi_cmd failed at byte %d\n", i);
	  return status;
	}
    }
  return SANE_STATUS_GOOD;
}


/* ----------------------------- PIE_USB_BULK_READ ------------------------------- */

static SANE_Status
pie_usb_bulk_read (SANE_Int dn, SANE_Byte * data, size_t len)
{
  SANE_Status status;
  size_t size;
  SANE_Byte outdata[8];

  DBG (DBG_proc, "pie_usb_bulk_read requesting %lu bytes\n", (u_long) len);

  if (len == 0)

    return SANE_STATUS_GOOD;

  memset (outdata, '\0', sizeof (outdata));

  while (len)
    {
      if (len > BULKIN_MAXSIZE)
	size = BULKIN_MAXSIZE;
      else
	size = len;

      outdata[4] = (size & 0xff);
      outdata[5] = ((size >> 8) & 0xff);
      outdata[6] = ((size >> 16) & 0xff);
      outdata[7] = ((size >> 24) & 0xff);

      status =
	sanei_usb_control_msg (dn, REQUEST_TYPE_OUT, REQUEST_BUFFER,
			       VALUE_BUFFER, INDEX, sizeof (outdata),
			       outdata);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (DBG_error,
	       "pie_usb_bulk_read failed while writing command: %s\n",
	       sane_strstatus (status));
	  return status;
	}

      DBG (DBG_info,
	   "pie_usb_bulk_read trying to read %lu bytes of data\n",
	   (u_long) size);
      status = sanei_usb_read_bulk (dn, data, &size);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (DBG_error,
	       "pie_usb_bulk_read failed while reading bulk data: %s\n",
	       sane_strstatus (status));
	  return status;
	}

      DBG (DBG_info,
	   "pie_usb_bulk_read read %lu bytes, %lu remaining\n",
	   (u_long) size, (u_long) (len - size));
      len -= size;
      data += size;
    }

  DBG (DBG_info, "pie_usb_bulk_read completed\n");
  return SANE_STATUS_GOOD;
}


/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ SCSI over USB ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ------------------------------- PIE_USB_READ ---------------------------------- */

/* note: the transfer length in a real 6-byte READ command can have two meanings:  */
/* Normally it means length in bytes, however, when an image is read it amounts to */
/* (lines * color-planes)                                                          */

static SANE_Status
pie_usb_read (int dn, const SANE_Byte * cmnd, void *buf, size_t * buf_len)
{
  SANE_Status status;
  size_t length = *buf_len;

  DBG (DBG_proc, "pie_usb_read\n");

  status = pie_usb_write_scsi_cmd (dn, cmnd);
  if (status != SANE_STATUS_GOOD)
    return status;
  status = pie_usb_ready_state (dn);
  if (status != SANE_STATUS_GOOD)
    return status;

  status = pie_usb_bulk_read (dn, buf, length);
  if (status != SANE_STATUS_GOOD)
    return status;

  return pie_usb_ready_state (dn);
}

/* ------------------------------- PIE_USB_WRITE --------------------------------- */

static SANE_Status
pie_usb_write (int dn, const SANE_Byte * cmnd, size_t length)
{
  SANE_Status status;
  SANE_Byte mnd;
  size_t i;

  DBG (DBG_info, "pie_usb_write\n");
  if (length <= 6)
    return SANE_STATUS_GOOD;

  status = pie_usb_write_scsi_cmd (dn, cmnd);
  if (status != SANE_STATUS_GOOD)
    return status;
  status = pie_usb_ready_state (dn);
  if (status != SANE_STATUS_GOOD)
    return status;

  DBG (DBG_info, "pie_usb_write: now writing %lu bytes\n",
       (u_long) length - 6);
  for (i = 6; i < length; i++)
    {
      mnd = cmnd[i];
      status =
	sanei_usb_control_msg (dn, REQUEST_TYPE_OUT, REQUEST_REGISTER,
			       VALUE_WRITE_REGISTER, INDEX, 1, &mnd);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (DBG_error, "pie_usb_write failed at byte %lu\n",
	       (unsigned long) length);
	  return status;
	}
    }

  return pie_usb_ready_state (dn);
}


/* ------------------------------ PIE_USB_COMMAND -------------------------------- */

static SANE_Status
pie_usb_command (int dn, const SANE_Byte * cmnd)
{
  SANE_Status status;

  DBG (DBG_info, "pie_usb_command\n");

  status = pie_usb_write_scsi_cmd (dn, cmnd);
  if (status != SANE_STATUS_GOOD)
    return status;

  return pie_usb_ready_state (dn);
}

/* ---------------------------- PIE_USB_SCSI_WRAPPER ----------------------------- */

/* Entry for SCSI over USB command wrapping                                        */

SANE_Status
pie_usb_scsi_wrapper (int fd, const void *src, size_t src_size, void *dst,
		      size_t * dst_size)
{
  SANE_Status status = SANE_STATUS_GOOD;
  const SANE_Byte *cmnd = src;

  if (cmnd[0] == INQUIRY)
    {
      status = pie_usb_write_control_sequence (fd, PIE_USB_Init_Sequence_1);
      if (status != SANE_STATUS_GOOD)
	return status;
      status = pie_usb_write_control_sequence (fd, PIE_USB_Init_Sequence_2);
      if (status != SANE_STATUS_GOOD)
	return status;
    }
  status = pie_usb_write_control_sequence (fd, PIE_USB_Setup_SCSI_Sequence);
  if (status != SANE_STATUS_GOOD)
    return status;

  switch (cmnd[0])
    {
    case TEST_UNIT_READY:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing TEST_UNIT_READY\n");
      status = pie_usb_command (fd, cmnd);
      break;
    case REQUEST_SENSE:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing REQUEST_SENSE\n");
      status = pie_usb_read (fd, cmnd, dst, dst_size);
      break;
    case READ:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing READ\n");
      status = pie_usb_read (fd, cmnd, dst, dst_size);
      break;
    case WRITE:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing WRITE\n");
      status = pie_usb_write (fd, cmnd, src_size);
      break;
    case INQUIRY:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing INQUIRY\n");
      status = pie_usb_read (fd, cmnd, dst, dst_size);
      break;
    case PARAM:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing PARAM\n");
      status = pie_usb_read (fd, cmnd, dst, dst_size);
      break;
    case MODE:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing MODE\n");
      status = pie_usb_write (fd, cmnd, src_size);
      break;
    case RESERVE_UNIT:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing RESERVE_UNIT\n");
      status = pie_usb_command (fd, cmnd);
      break;
    case RELEASE_UNIT:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing RELEASE_UNIT\n");
      status = pie_usb_command (fd, cmnd);
      break;
    case PIE_COPY:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing PIE_COPY\n");
      status = pie_usb_read (fd, cmnd, dst, dst_size);
      break;
    case SCAN:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing SCAN\n");
      status = pie_usb_command (fd, cmnd);
      break;
    case PIE_IMAGE_CLOSE:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing PIE_IMAGE_CLOSE\n");
      status = pie_usb_write (fd, cmnd, src_size);
      break;
    case PIE_READ_CALIBRATION:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing PIE_READ_CALIBRATION\n");
      status = pie_usb_read (fd, cmnd, dst, dst_size);
      break;
    case PIE_WRITE_CALIBRATION:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing PIE_WRITE_CALIBRATION\n");
      status = pie_usb_write (fd, cmnd, src_size);
      break;
    case PIE_READ_STATUS:
      DBG (DBG_info, "pie_usb_scsi_wrapper doing PIE_READ_STATUS\n");
      status = pie_usb_read (fd, cmnd, dst, dst_size);
      break;
    default:
      DBG (DBG_info, "pie_usb_scsi_wrapper failed for command 0x%02x\n",
	   cmnd[0]);
      status = SANE_STATUS_INVAL;
      break;
    }

  return status;
}


/* ---------------------------- PIE_USB_REQUEST_SENSE ---------------------------- */

static SANE_Status
pie_usb_request_sense (int dn, uint32_t * kascq)
{
  unsigned char buffer[16];	/* for PIE's sense */
  size_t size;
  SANE_Status status;

  DBG (DBG_proc, "pie_usb_request_sense\n");

  size = 14;
  set_RS_allocation_length (request_senseC, size);

  status =
    pie_usb_scsi_wrapper (dn, request_senseC, sizeof (request_senseC), buffer,
			  &size);
  if (status != SANE_STATUS_GOOD)
    {
      DBG (DBG_error, "pie_usb_request_sense failed\n");
      return status;
    }
  *kascq = ((int) get_RS_sense_key (buffer) << 16) +
    ((int) get_RS_ASC (buffer) << 8) + (int) get_RS_ASCQ (buffer);
  sense_handler (dn, buffer, NULL);

  return SANE_STATUS_GOOD;
}


/* -------------------------- PIE_USB_CLOSE_IMAGE_READ --------------------------- */

/* PIE vendor specific, called after image aquisition, waits for ready */

static SANE_Status
pie_usb_close_image_read (Pie_Scanner * scanner)
{
  unsigned char buffer[16];
  unsigned char *data;
  SANE_Status status;
  size_t size = image_closeC[4];
  int cnt = 0;

  DBG (DBG_proc, "pie_usb_close_image_read\n");

  memcpy (buffer, image_closeC, sizeof (image_closeC));

  data = buffer + sizeof (image_closeC);
  memset (data, 0, size);
  data[0] = 2;
  size += sizeof (image_closeC);

  do
    {
      status = pie_usb_scsi_wrapper (scanner->sfd, buffer, size, NULL, NULL);

      if (status == SANE_STATUS_GOOD)
	return status;

      if (cnt == 1)
	DBG (DBG_info2,
	     "pie_usb_close_image_read: scanner reports %s, waiting ...\n",
	     sane_strstatus (status));

      usleep (TUR_WAIT_TIME);
      cnt++;
    }
  while (cnt < 20);		/* maximally 20 * 0.5 = 10 sec */

  return status;
}

/* --------------------------- PIE_USB_READ_STATUS ---------------------------- */

/* PIE vendor specific 0xdd, called during idle, 
   initial internal calibration, ...,  reads e.g. button state */

static SANE_Status
pie_usb_read_status (int dn, unsigned char *buf)
{
  SANE_Status status;
  size_t size;

  DBG (DBG_proc, "pie_usb_read_status\n");

  size = 11;

  status =
    pie_usb_scsi_wrapper (dn, read_statusC, sizeof (read_statusC), buf,
			  &size);
  return status;
}


/* ---------------------------- PIE_USB_WAIT_SCANNER ----------------------------- */

static SANE_Status
pie_usb_wait_scanner (Pie_Scanner * scanner, int secs)
{
  SANE_Status status;
  int cnt = secs * 16;

  DBG (DBG_proc, "pie_usb_wait_scanner\n");

  do
    {
      status = pie_usb_scsi_wrapper (scanner->sfd, test_unit_ready.cmd,
				     test_unit_ready.size, NULL, NULL);
      if (status == SANE_STATUS_GOOD)
	return status;
      if (cnt == 0)
	{
	  DBG (DBG_warning, "pie_usb_wait_scanner timed out\n");
	  return status;
	}
      usleep (62500);
      cnt--;

    }
  while (status == SANE_STATUS_DEVICE_BUSY);

  DBG (DBG_error,
       "pie_usb_wait_scanner failed: %s\n", sane_strstatus (status));
  return status;
}


/* -------------------------- PIE_USB_COPY_SENSORS ---------------------------- */

/* pie_usb_copy_sensors reads a vector consisting of 0x00 and 0x70 before image
   aquisition. 0x00 seems to indicate that the sensor element is used for the
   following image. This vector is used for shading correction in software. */

static SANE_Status
pie_usb_copy_sensors (Pie_Scanner * scanner)
{
  size_t size_read;
  int cnt = 0;
  SANE_Status status;

  DBG (DBG_proc, "pie_usb_copy_sensors\n");

  do
    {
      size_read = scanner->device->cal_info[0].pixels_per_line;
      status =
	pie_usb_scsi_wrapper (scanner->sfd, pie_copyC, sizeof (pie_copyC),
			      scanner->cal_data->sensors, &size_read);
      if (status == SANE_STATUS_GOOD)
	return status;

      if (cnt == 1)
	DBG (DBG_info2,
	     "pie_usb_copy_sensors: scanner reports %s, waiting ...\n",
	     sane_strstatus (status));

      usleep (TUR_WAIT_TIME);
      cnt++;
    }
  while (cnt < 10);		/* maximally 10 * 0.5 = 5 sec */

  return status;
}

/* ----------------------------- PIE_USB_ATTACH_OPEN ----------------------------- */

/* called during initialization, tries to open and identify an USB scanner	   */

static SANE_Status
pie_usb_attach_open (SANE_String_Const devname, SANE_Int * dn)
{
  SANE_Status status;
  SANE_Int vendor, product;
  SANE_Int model = 0;
  int i;

  DBG (DBG_proc, "pie_usb_attach_open: opening `%s'\n", devname);
  status = sanei_usb_open (devname, dn);
  if (status != SANE_STATUS_GOOD)
    {
      DBG (DBG_error, "pie_usb_attach_open: sanei_usb_open failed\n");
      return status;
    }
  DBG (DBG_info, "pie_usb_attach_open: USB device `%s' successfully opened\n",
       devname);

  status = sanei_usb_get_vendor_product (*dn, &vendor, &product);
  if (status != SANE_STATUS_GOOD)
    {
      DBG (DBG_error,
	   "pie_usb_attach_open: couldn't get vendor and product ids of device `%s': %s\n",
	   devname, sane_strstatus (status));
      return status;
    }

  for (i = 0; pie_usb_device_list[i].model != 0; i++)
    {
      if (vendor == pie_usb_device_list[i].vendor &&
	  product == pie_usb_device_list[i].product)
	{
	  model++;
	  break;
	}
    }
  if (model == 0)
    {
      DBG (DBG_error,
	   "pie_usb_attach_open: vendor 0x%04x product 0x%04x is not supported by this backend\n",
	   vendor, product);
      status = SANE_STATUS_INVAL;
    }

  return status;
}



/* -------------------------------- TRY_ATTACH_USB ------------------------------- */

/* callback function for sanei_config_attach_matching_devices which has checked    */
/* for SCSI scanners. Here we do it for USB ones                                   */

static SANE_Status
try_attach_usb (const char *name)
{
  sanei_usb_attach_matching_devices (name, attach_one);
  return SANE_STATUS_GOOD;
}


/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ calibration functions ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ------------------------ PIE_USB_CALICALC_HIQUAL ----------------------------*/

/* Empirical data show that in quality mode for a color (j) the illumination (Ij)
   reached with certain gain (gj) and exposure time (tj) settings can be described
   by (I)  Ij = aj * exp (c * gj^2) * tj . The gain constants gj and desired brightness
   has to be provided. All other values can be derived from the calibration lines
   or the 0xd7 read. In quality mode maximal gain is limited and most of the
   calibration is done with exposure time  */

static void
pie_usb_calicalc_hiqual (Pie_Scanner * scanner,
			 PIE_USB_Calibration_Read * d7cal)
{
  PIE_USB_Model *model = scanner->device->model;
  PIE_USB_Calibration *caldat = scanner->cal_data;
  double fact, dgain;
  int i, tg, tt, tmax;

  SANE_Status status;
  int brightnesses[3];
  int pokebuf[4];
  int pokesiz = 3;

  DBG (DBG_proc, "pie_usb_calicalc_hiqual\n");

  for (i = 0; i < 3; i++)
    brightnesses[i] = caldat->brightness[i];
  status = pie_usb_poke_decints ("/tmp/bright.txt", pokebuf, &pokesiz);
  if (status == SANE_STATUS_GOOD && pokesiz)
    {
      for (i = 0; i < 3; i++)
	brightnesses[i] = pokebuf[i];
      DBG (DBG_info,
	   "pie_usb_calicalc_hiqual poked brightness %d, %d, %d\n",
	   brightnesses[0], brightnesses[1], brightnesses[2]);
    }

  tmax = 0;
  for (i = 0; i < 3; i++)
    {
      /* overall illumination correction factor
         fact = I / I0 = exp (c * (g^2 - g0^2)) * t / t0 */
      fact = (double) brightnesses[i] / (double) caldat->mean_shade[i];
      /* calculate gain from f^(1/p) = exp (c * (g^2 - g0^2)), the part done by gain */
      dgain =
	log (fact) / (model->gain_const[i] *
		      (double) model->gain_hiqual_part) +
	(double) d7cal->gain[i] * (double) d7cal->gain[i];
      if (dgain < 0)
	tg = model->gain_min;
      else
	{
	  tg = sqrt (dgain) + 0.5;
	  if (tg < model->gain_min)
	    tg = model->gain_min;
	  if (tg > model->gain_hiqual_max)
	    tg = model->gain_hiqual_max;
	}
      caldat->cal_hiqual.gain[i] = tg;
      DBG (DBG_info,
	   "pie_usb_calicalc_hiqual gain[%d] = 0x%02x = %d\n", i, tg, tg);

      /* the rest has to be done by exposure time */
      dgain = fact;
      if (tg != d7cal->gain[i])
	dgain *=
	  exp (model->gain_const[i] *
	       ((double) d7cal->gain[i] * (double) d7cal->gain[i] -
		(double) tg * (double) tg));
      tt = dgain * (double) d7cal->texp[i] + 0.5;

      /* if exposure time is too short try to redo the gain,
         should not happen very often as "fact" is usually > 1 */
      if (tt < d7cal->t_min)
	{
	  tt = d7cal->t_min;
	  fact *= (double) tt / (double) d7cal->texp[i];
	  dgain =
	    log (fact) / model->gain_const[i] +
	    (double) d7cal->gain[i] * (double) d7cal->gain[i];
	  if (dgain < 0)
	    tg = model->gain_min;
	  else
	    {
	      tg = sqrt (dgain) + 0.5;
	      if (tg < model->gain_min)
		tg = model->gain_min;
	    }
	  caldat->cal_hiqual.gain[i] = tg;
	  DBG (DBG_info,
	       "pie_usb_calicalc_hiqual regain[%d] = 0x%02x = %d\n",
	       i, tg, tg);
	}
      caldat->cal_hiqual.texp[i] = tt;
      if (tt > tmax)
	tmax = tt;
      DBG (DBG_info,
	   "pie_usb_calicalc_hiqual texp[%d] = 0x%02x = %d\n", i, tt, tt);
    }
  caldat->cal_hiqual.texp_max = tmax;
}

/* ------------------------ PIE_USB_CALICALC_NORMAL ----------------------------*/

/* In normal mode we have to fight nasty offsets with some types of scanners and
   a rather narrow window for exposure time. Here for a color (j) the illumination
   (Ij) reached with certain gain (gj) and exposure time (tj) settings can be described
   by (II)  I = n * aj * exp(c * gj^2) * tj - (n - 1) * 65536 . The gain constant c, offset
   factor n and desired brightness have to be provided. Gain may be really maximal here,
   i.e. 0x3f, the exposure time has to be larger than t_min */

static void
pie_usb_calicalc_normal (Pie_Scanner * scanner,
			 PIE_USB_Calibration_Read * d7cal)
{
  PIE_USB_Model *model = scanner->device->model;
  PIE_USB_Calibration *caldat = scanner->cal_data;
  int tg, tt, i, tmax;

  SANE_Status status;
  int brightnesses[3];
  int pokebuf[4];
  int pokesiz = 3;

  DBG (DBG_proc, "pie_usb_calicalc_normal\n");

  for (i = 0; i < 3; i++)
    brightnesses[i] = caldat->brightness[i];
  status = pie_usb_poke_decints ("/tmp/bright.txt", pokebuf, &pokesiz);
  if (status == SANE_STATUS_GOOD && pokesiz)
    {
      for (i = 0; i < 3; i++)
	brightnesses[i] = pokebuf[i];
      DBG (DBG_info,
	   "pie_usb_calicalc_normal poked brightness %d, %d, %d\n",
	   brightnesses[0], brightnesses[1], brightnesses[2]);
    }

  tmax = 0;
  for (i = 0; i < 3; i++)
    {
      /* start calculating gain for default texp rounding down.
         Following horrible expression is obtained by setting up equation (II) for the desired
         brightness, setting up quality mode equation (I) using the information from the
         0xd7 read together with the achieved brightness of read calibration lines and solving
         that system for the gain required in normal mode */
      tg = sqrt (log ((((double) brightnesses[i] +
			((model->offs_factor[i] -
			  1.0) * 65536)) * (double) d7cal->texp[i]) /
		      (model->offs_factor[i] *
		       (double) model->default_normal.texp[i] *
		       (double) caldat->mean_shade[i])) /
		 model->gain_const[i] +
		 (double) d7cal->gain[i] * (double) d7cal->gain[i]);
      if (tg > 0x3f)
	tg = 0x3f;
      /* here the gain (tg) is smaller than for what it was calculated, so we need
         a longer exposure time. Using the same set of equations as above exposure time is
         calculated for the given gain */
      tt = (((double) brightnesses[i] + ((model->offs_factor[i] -
					  1.0) * 65536.0)) *
	    (double) d7cal->texp[i]) / (model->offs_factor[i] *
					(double) caldat->mean_shade[i]) *
	exp (model->gain_const[i] *
	     ((double) d7cal->gain[i] * (double) d7cal->gain[i] - tg * tg)) +
	0.5;
      if (tt < d7cal->t_min)
	tt = d7cal->t_min;
      if (tt > model->texp_normal_max)
	tt = model->texp_normal_max;

      caldat->cal_normal.gain[i] = tg;
      caldat->cal_normal.texp[i] = tt;
      if (tt > tmax)
	tmax = tt;
      DBG (DBG_info,
	   "pie_usb_calicalc_normal gain[%d] = 0x%02x = %d, texp[%d] = 0x%02x = %d\n",
	   i, tg, tg, i, tt, tt);
    }
  caldat->cal_normal.texp_max = tmax;
}


/* ---------------------- PIE_USB_CALICALC_SLOW_DOWN ---------------------------*/

/* at high resolutions to much data seem to be delivered to the scanners' USB
   interface. So we have to slow down. As we do not have the results from the
   PARAMS call yet calculate from our internal values */

static unsigned char
pie_usb_calicalc_slow_down (Pie_Scanner * scanner)
{
  double width, x_dpmm;
  int slow_down;

  DBG (DBG_proc, "pie_usb_calicalc_slow_down\n");

  slow_down = 0;
  width = SANE_UNFIX (scanner->val[OPT_BR_X].w - scanner->val[OPT_TL_X].w);
  x_dpmm = SANE_UNFIX (scanner->val[OPT_RESOLUTION].w) / MM_PER_INCH;
  if ((width > 0) && (x_dpmm > 0))
    {
      width *= 3 * x_dpmm;		/* calculate for RGB */
      if (scanner->val[OPT_BIT_DEPTH].w > 8)
	width *= 2;			/* bytes per line */

      if (width > SLOW_START)
	{				/* slow down value */
	  width = (width - SLOW_START) * SLOW_HEIGHT / SLOW_LENGTH;
	  if (scanner->cal_mode ==		/* correct for exposure time */
	      scanner->device->model->op_mode[OPM_QUALITY])
	    width *=
	      (double) scanner->device->model->default_hiqual.texp_max /
	      (double) scanner->cal_data->cal_hiqual.texp_max;
	  else
	    width *=
	      (double) scanner->device->model->default_normal.texp_max /
	      (double) scanner->cal_data->cal_normal.texp_max;
	  slow_down = (int) (width + 0.5);
	  if (scanner->colormode == RGBI)	/* correct for infrared */
	    slow_down -= 2;
	  if ((slow_down < 0) || (slow_down >= 16))
	    slow_down = 0;
	}
    }
  DBG (DBG_info, "pie_usb_calicalc_slow_down: %d\n", slow_down);
  return slow_down;
}


/* ------------------------ PIE_USB_SEND_CALIBRATION -------------------------- */

/* This routine contains the vendor SCSI commands 0xd7 for reading and 0xdc
   for writing. As in the logs we first read 103 bytes and then write 23 bytes.
   only R G and B exposure times (texp) and gains are calibrated all other
   calibration values are copied from the 0xd7 read. */

static SANE_Status
pie_usb_send_calibration (Pie_Scanner * scanner, int calc_cal)
{
  PIE_USB_Calibration_Read calD7in;
  PIE_USB_Calibration_Send calDCout;
  size_t size_read = 103;
  size_t size_write = 23;
  SANE_Status status;
  unsigned char pokebuf[64];
  int i, pokesiz;

  DBG (DBG_proc, "pie_usb_send_calibration\n");

  status =			/* 103 bytes */
    pie_usb_scsi_wrapper (scanner->sfd, read_calibrationC,
			  sizeof (read_calibrationC), &calD7in, &size_read);
  if (status != SANE_STATUS_GOOD)
    return status;

  DBG (DBG_info, "pie_usb_send_calibration received:\n");
  DBG_DUMP (DBG_info, (unsigned char *) &calD7in, 103);

  if (calD7in.illumination[0] == 0)	/* assume same target as for blue */
    calD7in.illumination[0] = calD7in.illumination[2];
  if (calD7in.illumination[1] == 0)	/* assume same target as for blue */
    calD7in.illumination[1] = calD7in.illumination[2];
  for (i = 0; i < 3; i++)
    scanner->cal_data->target_shade[i] = calD7in.illumination[i];
  scanner->cal_data->target_shade[3] = calD7in.illumination[2];	/* ired ?? */

  /* now we have all data to calculate calibration */
  if (calc_cal)
    {
      pie_usb_calicalc_hiqual (scanner, &calD7in);
      pie_usb_calicalc_normal (scanner, &calD7in);
    }
  /* load vector to be sent */
  size_write += 6;
  memcpy (&(calDCout.scsi_cmd), write_calibrationC, 6);
  /* first with what we have/had calculated */
  if (scanner->cal_mode != scanner->device->model->op_mode[OPM_QUALITY])	/* normal mode */
    {
      for (i = 0; i < 3; i++)
	calDCout.texp[i] = scanner->cal_data->cal_normal.texp[i];
      for (i = 0; i < 3; i++)
	calDCout.gain[i] = scanner->cal_data->cal_normal.gain[i];
    }
  else				/* quality mode */
    {
      for (i = 0; i < 3; i++)
	calDCout.texp[i] = scanner->cal_data->cal_hiqual.texp[i];
      for (i = 0; i < 3; i++)
	calDCout.gain[i] = scanner->cal_data->cal_hiqual.gain[i];
    }

  /* then with values which are usully copied */
  memcpy (calDCout.offset, calD7in.offset, 6);	/* offsets, zero_2 */

  calDCout.some_time[0] = calD7in.some_time;
  /* slow down at high resolutions */
  if (calc_cal || (scanner->cal_mode == scanner->device->model->op_mode[OPM_SKIPCAL]))
    calDCout.some_time[1] = pie_usb_calicalc_slow_down (scanner);
  else
    calDCout.some_time[1] = 0;
  calDCout.some_time[2] = 0;

  memcpy (&calDCout.infrared, &calD7in.infrared, sizeof (PIE_USB_cal_ired));
  pokesiz = 12;
  status = pie_usb_poke_bytes ("/tmp/calbytes.txt", pokebuf, &pokesiz);
  if (status == SANE_STATUS_GOOD)
    {
      memcpy (calDCout.texp, &pokebuf[0], 6);
      memcpy (calDCout.gain, &pokebuf[6], 6);
    }

  DBG (DBG_info, "pie_usb_send_calibration sending:\n");
  DBG_DUMP (DBG_info, (unsigned char *) &calDCout, size_write);
  status =
    pie_usb_scsi_wrapper (scanner->sfd, &calDCout, size_write, NULL, NULL);
  return status;
}


/* ---------------------------- PIE_USB_CALIBRATE ----------------------------- */

/* 1) 1 line of image data is read
   2) test unit ready
   4)  13 lines are read  in indexed RGBI
   5)  something should be learned from that
   6)  calibration info is sent to the scanner
   7) dump left over lines
   scanner->device->cal_info[any].num_lines contains 45
*/

static SANE_Status
pie_usb_calibrate (Pie_Scanner * scanner)
{
  unsigned int val;
  unsigned char *rcv_buffer, *src;
  int rcv_length, rcv_lines, rcv_bits;
  int pixels_per_line, average_lines;
  int i, j, k, l;
  int n[4];
  int *dest;
  double dval;
  size_t size;
  SANE_Status status;
  DBG (DBG_proc, "pie_usb_calibrate\n");
  rcv_lines = scanner->device->cal_info[0].num_lines;
  average_lines = rcv_lines - 1;
  if (AVERAGE_CAL_LINES < average_lines)
    average_lines = AVERAGE_CAL_LINES;
  pixels_per_line = scanner->device->cal_info[0].pixels_per_line;
  rcv_length = pixels_per_line;
  rcv_bits = scanner->device->cal_info[0].receive_bits;
  if (rcv_bits > 8)
    rcv_length *= 2;
  if ((scanner->colormode == RGB) || (scanner->colormode == RGBI))
    rcv_length = (rcv_length + 2) * 4;
  else
    return SANE_STATUS_INVAL;
  rcv_buffer = (unsigned char *) malloc (rcv_length);
  if (!rcv_buffer)
    return SANE_STATUS_NO_MEM;
  status = pie_wait_scanner (scanner);
  if (status != SANE_STATUS_GOOD)
    goto freend;
  set_read_length (sread.cmd, 4);	/* a test line */
  size = rcv_length;
  status =
    pie_usb_scsi_wrapper (scanner->sfd, sread.cmd, sread.size,
			  rcv_buffer, &size);
  if (status != SANE_STATUS_GOOD)	/* should not happen */
    {
      status = SANE_STATUS_GOOD;	/* try to recover */
      goto freend;
    }
  rcv_lines--;
  /* this gets the data */
  memset (scanner->cal_data->shades, 0, pixels_per_line * 4 * sizeof (int));
  scanner->cal_data->shades[0] = 0;
  for (i = 0; i < 4; i++)
    n[i] = 0;
  for (k = 0; k < average_lines; k++)
    {
      size = rcv_length;
      status =
	pie_usb_scsi_wrapper (scanner->sfd, sread.cmd, sread.size,
			      rcv_buffer, &size);
      if (status != SANE_STATUS_GOOD)
	goto freend;
      for (i = 0; i < 4; i++)
	{
	  if (rcv_bits > 8)
	    src = rcv_buffer + i * (pixels_per_line + 1) * 2;
	  else
	    src = rcv_buffer + i * (pixels_per_line + 2);
	  if (*src == 'R')
	    l = 0;
	  else if (*src == 'G')
	    l = 1;
	  else if (*src == 'B')
	    l = 2;
	  else if (*src == 'I')
	    l = 3;
	  else
	    {
	      DBG (DBG_error,
		   "pie_usb_calibrate: invalid index byte (%02x)\n", *src);
	      DBG_DUMP (DBG_error, src, 32);
	      status = SANE_STATUS_INVAL;
	      goto freend;
	    }
	  src += 2;
	  dest = scanner->cal_data->shades + l * pixels_per_line;
	  if (rcv_bits > 8)
	    {
	      for (j = 0; j < pixels_per_line; j++)
		{
		  val = *src++;
		  val += (*src++) << 8;
		  *dest++ += val;
		}
	    }
	  else
	    {
	      for (j = 0; j < pixels_per_line; j++)
		*dest++ += *src++;
	    }
	  n[l]++;
	}
      rcv_lines--;
    }
  for (l = 0; l < 4; l++)
    {
      dest = scanner->cal_data->shades + l * pixels_per_line;
      dval = 0;
      for (j = pixels_per_line; j > 0; j--)
	{
	  dval += *dest;
	  *dest++ /= n[l];
	}
      scanner->cal_data->mean_shade[l] =
	dval / (double) (n[l] * pixels_per_line) + 0.5;
      DBG (DBG_info,
	   "pie_usb_calibrate: color %d, mean %d = %04x\n", l,
	   scanner->cal_data->mean_shade[l],
	   scanner->cal_data->mean_shade[l]);
    }
  if (DBG_LEVEL >= DBG_dump)
    pie_usb_shades_to_pnm (scanner, "/tmp/pieshading", 64);
  /* get, calculate and send calibration */
  status = pie_usb_send_calibration (scanner, 1);
  if (status != SANE_STATUS_GOOD)
    goto freend;
  /* we have to read all lines, discard the rest */
  while (rcv_lines > 0)
    {
      size = rcv_length;
      status =
	pie_usb_scsi_wrapper (scanner->sfd, sread.cmd, sread.size,
			      rcv_buffer, &size);
      if (status != SANE_STATUS_GOOD)
	goto freend;
      rcv_lines--;
    }

freend:			/* ugly, but this may be one of the cases ... */
  free (rcv_buffer);
  return status;
}


/* ------------------------- PIE_USB_CORRECT_SHADING -------------------------- */

static void
pie_usb_correct_shading (Pie_Scanner * scanner,
			 unsigned char *buffer, int pixels,
			 int bits, int start_plane, int end_plane)
{
  unsigned char *sensors, *buf;
  unsigned int val;
  int *shade[4];
  int target[4];
  int cal_pixels;
  int cal_idx = 0;
  int i, j, k;

  DBG (DBG_proc, "pie_usb_correct_shading: %d to %d\n", start_plane, end_plane);

  cal_pixels = scanner->device->cal_info[0].pixels_per_line;
  sensors = scanner->cal_data->sensors;
  for (k = start_plane; k <= end_plane; k++)
    {
      shade[k] = scanner->cal_data->shades + k * cal_pixels;
      /* we can take means or targets here,
         targets may lead to values larger than 16 bit */
      target[k] = scanner->cal_data->mean_shade[k];
    }
  buf = buffer;
  j = 0;
  for (i = 0; i < pixels; i++)
    {
      while (sensors[j] != 0 && j < cal_pixels)
	j++;
      if (j < cal_pixels)
	{
	  cal_idx = j;
	  j++;
	}
      if (bits > 8)
	for (k = start_plane; k <= end_plane; k++)
	  {
	    val = *buf++;
	    val += (*buf--) << 8;
	    val = (val * target[k]) / shade[k][cal_idx];
	    if (val > 0xffff)
	      val = 0xffff;
	    *buf++ = val & 0xff;
	    *buf++ = val >> 8;
	  }
      else
	for (k = start_plane; k <= end_plane; k++)
	  {
	    val = *buf;
	    val = (val * target[k]) / shade[k][cal_idx];
	    if (val > 0xff)
	      val = 0xff;
	    *buf++ = val;
	  }
    }
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ other high level functions ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ------------------------ PIE_USB_IS_SCANNER_READY -------------------------- */

static SANE_Status
pie_usb_is_scanner_ready (int dn)
{
  unsigned char buffer[16];
  SANE_Status status, state;
  uint32_t sense_kascq;
  int wait_cnt = 240;		/* 240 * 0.5 = 120 seconds */
  DBG (DBG_proc, "pie_usb_is_scanner_ready\n");
  do
    {
      status =
	pie_usb_scsi_wrapper (dn, test_unit_ready.cmd,
			      test_unit_ready.size, NULL, NULL);
      if (status == SANE_STATUS_IO_ERROR)	/* Not Ready - Warming Up ? */
	{
	  state = pie_usb_request_sense (dn, &sense_kascq);
	  if (state != SANE_STATUS_GOOD)
	    return state;
	  if (sense_kascq != 0x020401)
	    return status;
	  else
	    status = SANE_STATUS_DEVICE_BUSY;
	}

      if (status == SANE_STATUS_DEVICE_BUSY)
	{
	  usleep (TUR_WAIT_TIME);
	  wait_cnt--;
	}

      memset (buffer, 0, 11);
      status = pie_usb_read_status (dn, buffer);
      if (status == SANE_STATUS_IO_ERROR)	/* Not Ready - Warming Up ? */
	{
	  state = pie_usb_request_sense (dn, &sense_kascq);
	  if (state != SANE_STATUS_GOOD)
	    return state;
	  if (sense_kascq != 0x020401)
	    return status;
	  else
	    status = SANE_STATUS_DEVICE_BUSY;
	}

      DBG_DUMP (DBG_info, buffer, 11);
      if ((buffer[9] != 0) || (buffer[10] != 0))
	status = SANE_STATUS_DEVICE_BUSY;
      if (status == SANE_STATUS_DEVICE_BUSY)
	{
	  usleep (TUR_WAIT_TIME);
	  wait_cnt--;
	}
    }
  while ((status == SANE_STATUS_DEVICE_BUSY) && (wait_cnt > 0));
  return status;
}


/* ------------------------------ PIE_USB_SCAN -------------------------------- */

static SANE_Status
pie_usb_scan (Pie_Scanner * scanner, int start)
{
  SANE_Status status, state;
  uint32_t sense_kascq;
  DBG (DBG_proc, "pie_usb_scan: %d\n", start);
  set_scan_cmd (scan.cmd, start);
  if (start)
    {
      /* wait upto X seconds until returned to start position */
      status = pie_usb_wait_scanner (scanner, 15);
      if (status != SANE_STATUS_GOOD)
	return status;
      do
	{
	  status =
	    pie_usb_scsi_wrapper (scanner->sfd, scan.cmd, scan.size,
				  NULL, NULL);
	  if (status)
	    {
	      DBG (DBG_error, "pie_usb_scan: received %s\n",
		   sane_strstatus (status));
	      if (status == SANE_STATUS_IO_ERROR)
		{
		  state = pie_usb_request_sense (scanner->sfd, &sense_kascq);
		  if (state != SANE_STATUS_GOOD)
		    return state;
		  if (sense_kascq == 0x020401)	/* Not Ready - Warming Up */
		    status = SANE_STATUS_DEVICE_BUSY;
		  else if (sense_kascq == 0x068200)	/* calibration disable not granted */
		    status = SANE_STATUS_GOOD;
		}
	      if (status == SANE_STATUS_DEVICE_BUSY)
		usleep (SCAN_WARMUP_WAIT_TIME);
	    }
	}
      while (status == SANE_STATUS_DEVICE_BUSY);
      return status;
    }
  else
    {
      status =
	pie_usb_scsi_wrapper (scanner->sfd, scan.cmd, scan.size, NULL, NULL);
      if (status == SANE_STATUS_IO_ERROR)
	{
	  state = pie_usb_request_sense (scanner->sfd, &sense_kascq);
	  if (state != SANE_STATUS_GOOD)
	    return state;
	  if (sense_kascq != 0x0b0006)	/* ABORT message from initiator */
	    return status;
	}
      return SANE_STATUS_GOOD;
    }
}


/*-------------------------- PIE_USB_MODE SELECT -------------------------------*/

static SANE_Status
pie_usb_mode_select (Pie_Scanner * scanner)
{
  PIE_USB_Model *model = scanner->device->model;
  SANE_Status status;
  unsigned char buffer[128];
  size_t size;
  unsigned char *data;
  int i, cal;
  DBG (DBG_proc, "pie_usb_mode_select\n");
  size = 16;
  set_mode_length (smode.cmd, size);
  memcpy (buffer, smode.cmd, smode.size);
  data = buffer + smode.size;
  memset (data, 0, size);
  /* size of data */
  data[1] = size - 2;
  /* set resolution required */
  set_data (data, 2, scanner->resolution, 2);
  /* set color filter and color depth */
  data[4] = 0;
  scanner->cal_filter = 0;
  switch (scanner->colormode)
    {
    case RGBI:
      data[4] = INQ_FILTER_IRED;
      scanner->cal_filter = INQ_FILTER_IRED;
    case RGB:
      if (scanner->device->inquiry_filters & INQ_ONE_PASS_COLOR)
	{
	  data[4] |= INQ_ONE_PASS_COLOR;
	  scanner->cal_filter |= FILTER_RED | FILTER_GREEN | FILTER_BLUE;
	}
      else
	{
	  DBG (DBG_error,
	       "pie_usb_mode_select: support for multipass color not yet implemented\n");
	  return SANE_STATUS_UNSUPPORTED;
	}
      if (scanner->val[OPT_BIT_DEPTH].w == 16)
	data[5] = INQ_COLOR_DEPTH_16;
      else
	data[5] = INQ_COLOR_DEPTH_8;
      break;
    case GRAYSCALE:
      /* choose which filter to use for monochrome mode */
      if (scanner->device->inquiry_filters & INQ_FILTER_NEUTRAL)
	{
	  data[4] = FILTER_NEUTRAL;
	  scanner->cal_filter = FILTER_NEUTRAL;
	}
      else if (scanner->device->inquiry_filters & INQ_FILTER_GREEN)
	{
	  data[4] = FILTER_GREEN;
	  scanner->cal_filter = FILTER_GREEN;
	}
      else if (scanner->device->inquiry_filters & INQ_FILTER_RED)
	{
	  data[4] = FILTER_RED;
	  scanner->cal_filter = FILTER_RED;
	}
      else if (scanner->device->inquiry_filters & INQ_FILTER_BLUE)
	{
	  data[4] = FILTER_BLUE;
	  scanner->cal_filter = FILTER_BLUE;
	}
      else
	{
	  DBG (DBG_error,
	       "pie_usb_mode_select: scanner doesn't appear to support monochrome\n");
	  return SANE_STATUS_UNSUPPORTED;
	}

      data[5] = INQ_COLOR_DEPTH_8;
    }

  /* choose color packing method */
  if (scanner->device->inquiry_color_format & INQ_COLOR_FORMAT_INDEX)
    data[6] = INQ_COLOR_FORMAT_INDEX;
  else if (scanner->device->inquiry_color_format & INQ_COLOR_FORMAT_LINE)
    data[6] = INQ_COLOR_FORMAT_LINE;
  else
    {
      DBG (DBG_error,
	   "pie_usb_mode_select: support for pixel packing not yet implemented\n");
      return SANE_STATUS_UNSUPPORTED;
    }

  /* choose data format */
  if (scanner->device->inquiry_image_format & INQ_IMG_FMT_INTEL)
    data[8] = INQ_IMG_FMT_INTEL;
  else
    {
      DBG (DBG_error,
	   "pie_usb_mode_select: support for Motorola format not yet implemented\n");
      return SANE_STATUS_UNSUPPORTED;
    }

  /* set required calibration and quality */
  i = 0;
  while (scanner->device->speed_list[i] != NULL)
    {
      if (strcmp
	  (scanner->device->speed_list[i], scanner->val[OPT_SPEED].s) == 0)
	break;
      i++;
    }

  if (scanner->device->speed_list[i] == NULL)
    i = 0;			/* neither should happen */
  if (i > 2)
    i = 2;
  cal = i;
  if (cal == 1)
    {
      if (scanner->val[OPT_PREVIEW].w == SANE_TRUE)
	cal = model->op_mode[OPM_PREVIEW];	/*CAL_MODE_FILM_NORMAL; */
      else
	cal = model->op_mode[OPM_QUALITY];	/*CAL_MODE_FILM_HIQUAL; */
    }
  /* skip calibration if no quality in this or last scan */
/*  if ((cal == CAL_MODE_FILM_NORMAL) &&
      ((scanner->cal_mode & CAL_MODE_FILM_HIQUAL) == CAL_MODE_FILM_NORMAL)) */
  if ((cal != model->op_mode[OPM_QUALITY]) &&
      (scanner->cal_mode != model->op_mode[OPM_QUALITY]))
    cal = model->op_mode[OPM_SKIPCAL];
  data[9] = cal;
  scanner->cal_mode = cal;
  /* set required halftone pattern */
  i = 0;
  while (scanner->device->halftone_list[i] != NULL)
    {
      if (strcmp
	  (scanner->device->halftone_list[i],
	   scanner->val[OPT_HALFTONE_PATTERN].s) == 0)
	break;
      i++;
    }

  if (scanner->device->halftone_list[i] == NULL)
    data[12] = 0;		/* halftone pattern */
  else
    data[12] = i;
  data[13] = SANE_UNFIX (scanner->val[OPT_THRESHOLD].w) * 255 / 100;	/* lineart threshold */
  data[14] = 0x10;
  DBG (DBG_info, "pie_usb_mode_select: speed %02x\n", data[9]);
  DBG (DBG_info, "pie_usb_mode_select: halftone %d\n", data[12]);
  DBG (DBG_info, "pie_usb_mode_select: threshold %02x\n", data[13]);
  DBG (DBG_info, "pie_usb_mode_select sending:\n");
  DBG_DUMP (DBG_info, data, size);
  status =
    pie_usb_scsi_wrapper (scanner->sfd, buffer, smode.size + size,
			  NULL, NULL);
  if (status)
    {
      DBG (DBG_error,
	   "pie_usb_mode_select: write command returned status %s\n",
	   sane_strstatus (status));
    }

  return status;
}


/* --------------------- PIE_USB_READER_PROCESS_INDEXED ----------------------- */

static int
pie_usb_reader_process_indexed (Pie_Scanner * scanner, FILE * fp)
{
  char idx_char[4] = {
    'R', 'G', 'B', 'I'
  };
  char *idx_ptr[4];
  char *iend_ptr[4];
  int idx_buf[4];
  int idx_found;
  unsigned char *buffer = NULL;
  unsigned char *reorder = NULL;
  unsigned char *wrt_ptr, *end_ptr;
  unsigned char *dest, *src[4];
  int bytes_per_line, bytes_per_color;
  int chunk_lines, lines_todo;
  int read_lines, write_lines;
  int status;
  int irgb;			/* number of color planes = values per pixel */
  int irgb_out;			/* limit output as long as SANE disregards infrared */
  int request_data, i, j;
  size_t chunk_size, size;

  struct timeval start_time, end_time;
  long int elapsed_time;
  unsigned char *color_buf[4];
  unsigned char *color_ptr[4];
  char filename[64];

  DBG (DBG_read,
       "pie_usb_reader_process_indexed reading %d lines of %d bytes/line\n",
       scanner->params.lines, scanner->params.bytes_per_line);

  bytes_per_color = scanner->bytes_per_line + 2;
  switch (scanner->colormode)
    {
    case RGB:
      irgb = 3;
      irgb_out = 3;
      bytes_per_line = scanner->params.bytes_per_line + 6;
      break;
    case RGBI:
      irgb = 4;
#ifdef SANE_FRAME_RGBI
      irgb_out = 4;
      bytes_per_line = scanner->params.bytes_per_line + 8;
#else
      irgb_out = 3;
      bytes_per_line = scanner->params.bytes_per_line + bytes_per_color + 6;
#endif
      break;
    default:
      irgb = 1;
      irgb_out = 1;		/* assume filter bytes not present */
      bytes_per_line = scanner->params.bytes_per_line;
    }

  /* split the image read into reasonably sized chunks */
  chunk_lines = (scanner->params.lines + 7) / 8;
  chunk_size = chunk_lines * bytes_per_line;
  if (chunk_size > BUFFER_MAXSIZE)	/* hardware limitation ? */
    {
      chunk_lines = BUFFER_MAXSIZE / bytes_per_line;
      chunk_size = chunk_lines * bytes_per_line;
    }

  size = bytes_per_line *		/* deskewing needs a minimum */
    (scanner->filter_offset1 + (scanner->filter_offset2 + 3) * 2);
  if (chunk_size < size)
    {
      chunk_lines = (size + bytes_per_line - 1) / bytes_per_line;
      chunk_size = chunk_lines * bytes_per_line;
    }

  if (chunk_lines > scanner->params.lines)	/* not larger than image */
    {
      chunk_lines = scanner->params.lines;
      chunk_size = chunk_lines * bytes_per_line;
    }

  buffer = malloc (chunk_size * 2);
  reorder = malloc (bytes_per_line);
  if (!buffer || !reorder)
    {
      free (buffer);
      free (reorder);
      return SANE_STATUS_NO_MEM;
    }

  if (DBG_LEVEL >= DBG_analyse)
    {
      for (i = 0; i < irgb; i++)
	{
	  color_buf[i] =
	    malloc (scanner->bytes_per_line * scanner->params.lines);
	  if (!color_buf[i])
	    {
	      for (j = 0; j < i; j++)
		free (color_buf[j]);
	      return SANE_STATUS_NO_MEM;
	    }
	  color_ptr[i] = color_buf[i];
	}
    }
  gettimeofday (&start_time, NULL);

  /* read one buffer in advance */
  lines_todo = chunk_lines;
  set_read_length (sread.cmd, lines_todo * irgb);
  size = lines_todo * bytes_per_line;
  do
    {
      status =
	pie_usb_scsi_wrapper (scanner->sfd, sread.cmd, sread.size,
			      buffer, &size);
    }
  while (status);
  DBG_DUMP (DBG_dump, buffer, 32);
  write_lines = scanner->params.lines;
  read_lines = write_lines - lines_todo;
  wrt_ptr = buffer + chunk_size;
  end_ptr = wrt_ptr + chunk_size;
  for (i = 0; i < irgb; i++)
    {
      idx_ptr[i] = (char *) buffer;
      iend_ptr[i] = idx_ptr[i] + chunk_size;
      idx_buf[i] = 1;
      src[i] = NULL;
    }
  request_data = 0;
  idx_found = 0;
  while (write_lines)
    {
      if (irgb > 1)		/* RGB, RGBI cases */
	{
	  for (i = 0; i < irgb; i++)	/* find indices */
	    {
	      while (src[i] == NULL)
		{
		  if (*idx_ptr[i] == idx_char[i])
		    {
		      src[i] = (unsigned char *) idx_ptr[i] + 2;
		      idx_found++;
		    }
		  /* advance pointers unconditionally */
		  idx_ptr[i] += bytes_per_color;
		  if (idx_ptr[i] >= iend_ptr[i])
		    {
		      /* check for wrap */
		      if (idx_ptr[i] >= (char *) end_ptr)
			idx_ptr[i] = (char *) buffer;
		      /* maintain private "end of buffer" */
		      iend_ptr[i] = idx_ptr[i] + chunk_size;
		      idx_buf[i]--;
		      /* request buffer fill if necessary */
		      if (idx_buf[i] == 0)
			{
			  request_data = 1;
			  break;
			}
		    }
		}
	    }

	  if (idx_found == irgb)	/* success, reorder and write a line */
	    {
	      if (DBG_LEVEL >= DBG_analyse)
		for (i = 0; i < irgb; i++)
		  {
		    memcpy (color_ptr[i], src[i], scanner->bytes_per_line);
		    pie_usb_correct_shading (scanner, color_ptr[i],
					     scanner->params.pixels_per_line,
					     scanner->params.depth, i, i);
		    color_ptr[i] += scanner->bytes_per_line;
		  }

	      dest = reorder;
	      if (scanner->params.depth > 8)
		{
		  for (j = scanner->params.pixels_per_line; j > 0; j--)
		    for (i = 0; i < irgb_out; i++)
		      {
			*dest++ = *src[i]++;
			*dest++ = *src[i]++;
		      }
		}
	      else
		{
		  for (j = scanner->params.pixels_per_line; j > 0; j--)
		    for (i = 0; i < irgb_out; i++)
		      *dest++ = *src[i]++;
		}
	      pie_usb_correct_shading (scanner, reorder,
				       scanner->params.pixels_per_line,
				       scanner->params.depth, 0,
				       irgb_out - 1);
	      fwrite (reorder, 1, scanner->params.bytes_per_line, fp);
	      for (i = 0; i < irgb; i++)
		src[i] = NULL;
	      idx_found = 0;
	      write_lines--;
	    }
	}
      else			/* one color plane cases, write through */
	{
	  fwrite (idx_ptr[1], 1, size, fp);
	  idx_ptr[1] = (char *) wrt_ptr;
	  write_lines -= lines_todo;
	  if (write_lines > 0)
	    request_data = 1;
	}

      if (request_data)		/* read next data */
	{
	  if (read_lines)
	    {
	      lines_todo = chunk_lines;
	      if (lines_todo > read_lines)
		lines_todo = read_lines;
	      set_read_length (sread.cmd, lines_todo * irgb);
	      size = lines_todo * bytes_per_line;
	      do
		{
		  status =
		    pie_usb_scsi_wrapper (scanner->sfd, sread.cmd,
					  sread.size, wrt_ptr, &size);
		}
	      while (status);
	      DBG_DUMP (DBG_dump, wrt_ptr, 32);
	      read_lines -= lines_todo;
	      wrt_ptr += chunk_size;
	      if (wrt_ptr >= end_ptr)
		wrt_ptr = buffer;
	      for (i = 0; i < irgb; i++)
		idx_buf[i]++;
	      request_data = 0;
	    }
	  else if (write_lines)
	    {
	      DBG (DBG_error,
		   "pie_usb_reader_process_indexed: deskew failed for %d lines\n",
		   write_lines);
	      write_lines = 0;
	    }
	}
    }
  gettimeofday (&end_time, NULL);
  elapsed_time = ((end_time.tv_sec * 1000000 + end_time.tv_usec)
		  - (start_time.tv_sec * 1000000 +
		     start_time.tv_usec)) / 1000;
  DBG (DBG_read,
       "pie_usb_reader_process_indexed finished %d bytes in %ld ms\n",
       scanner->params.lines * scanner->params.bytes_per_line, elapsed_time);


  if (DBG_LEVEL >= DBG_analyse)
    {
      for (i = 0; i < irgb; i++)
	{
	  snprintf (filename, 63, "/tmp/color-%d.pnm", i);
	  pie_usb_write_pnm_file (filename, color_buf[i],
				  scanner->params.depth, 1,
				  scanner->params.pixels_per_line,
				  scanner->params.lines);
	}

      for (i = 0; i < irgb; i++)
	free (color_buf[i]);
    }

  free (buffer);
  free (reorder);
  return 0;
}


/* ------------------------- PIE_USB_READER_PROCESS --------------------------- */

static int
pie_usb_reader_process (Pie_Scanner * scanner, FILE * fp)
{
  int status;
  int lines, lines_todo, chunk_lines;
  unsigned char *buffer, *reorder;
  size_t size, chunk_size;
  DBG (DBG_read,
       "pie_usb_reader_process reading %d lines of %d bytes/line\n",
       scanner->params.lines, scanner->params.bytes_per_line);
  chunk_lines = (scanner->params.lines + 7) / 8;
  chunk_size = chunk_lines * scanner->params.bytes_per_line;
  if (chunk_size > BUFFER_MAXSIZE)
    {
      chunk_lines = BUFFER_MAXSIZE / scanner->params.bytes_per_line;
      chunk_size = chunk_lines * scanner->params.bytes_per_line;
    }
  buffer = malloc (chunk_size);
  reorder = malloc (chunk_size);
  if (!buffer || !reorder)
    {
      free (buffer);
      free (reorder);
      return SANE_STATUS_NO_MEM;
    }

  lines = scanner->params.lines;
  while (lines > 0)
    {
      lines_todo = chunk_lines;
      if (lines_todo > lines)
	lines_todo = lines;
      set_read_length (sread.cmd, lines_todo);
      size = lines_todo * scanner->params.bytes_per_line;
      do
	{
	  status =
	    pie_usb_scsi_wrapper (scanner->sfd, sread.cmd,
				  sread.size, buffer, &size);
	}
      while (status);
      DBG_DUMP (DBG_dump, buffer, 64);
      if (scanner->colormode == RGB)
	{
	  int offset = scanner->bytes_per_line;
	  unsigned char *dest = reorder;
	  unsigned char *src = buffer;
	  int i, j;
	  if (scanner->params.depth > 8)
	    {
	      for (j = lines_todo; j > 0; j--)
		{
		  for (i = scanner->params.pixels_per_line; i > 0; i--)
		    {
		      *dest++ = *src;
		      *dest++ = *(src + 1);
		      *dest++ = *(src + offset);
		      *dest++ = *(src + offset + 1);
		      *dest++ = *(src + 2 * offset);
		      *dest++ = *(src + 2 * offset + 1);
		      src++;
		      src++;
		    }
		  src += offset * 2;
		}
	    }
	  else
	    {
	      for (j = lines_todo; j > 0; j--)
		{
		  for (i = scanner->params.pixels_per_line; i > 0; i--)
		    {
		      *dest++ = *src;
		      *dest++ = *(src + offset);
		      *dest++ = *(src + 2 * offset);
		      src++;
		    }
		  src += offset * 2;
		}
	    }
	  fwrite (reorder, 1, size, fp);
	}
      else
	{
	  fwrite (buffer, 1, size, fp);
	}
      lines -= lines_todo;
    }

  free (buffer);
  free (reorder);
  return 0;
}

